#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <asm/param.h>
#include <asm/page.h>
#include <pwd.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>
#include <string.h>
#include <math.h>

#define HOSTS
#include "config.h"
#include "global.h"

#define F_NO_RESPONSE 1
#define F_ANSWERED    4


/*  global stuff */
int h_length;
int Hflags[NHOSTS];
int news=0;
uid_t myuid;

GMutex *mutex1;
GMutex *mutex2;

/* storage for %CPU calculation and  hardware info*/
struct p_tab {
    double time;
    int pid;
    long cputime;
};

struct pp_tab {
    int no;
    struct p_tab p[MAX_PROCS_PER_HOST];
};

static int pcpu[NHOSTS][MAX_PROCS_PER_HOST];
static int procno[NHOSTS][MAX_PROCS_PER_HOST];

static struct pp_tab Ptab[NHOSTS];

/* table header stuff */

struct header_field {
    gchar * name;
    GtkJustification Justification;
    gchar * tooltip;
};


#define Nfields  6
#define Nfields1 10
#define Nfields2 10
#define NfieldsMax 10

static struct header_field h_fields[Nfields] = {
    {" hostname",   GTK_JUSTIFY_LEFT,  NULL},
    {" Processors", GTK_JUSTIFY_LEFT,
     "PII = Pentium II, PPro = Pentium Pro, D = Deschutes, K = Klamath" },
    {"RAM",         GTK_JUSTIFY_CENTER,
     "total RAM in Mbyte"},
    {"uptime ",     GTK_JUSTIFY_RIGHT,  NULL},
    {"load ",       GTK_JUSTIFY_RIGHT,
    "one minute load average"},
    {"", GTK_JUSTIFY_CENTER, NULL}
};

static struct header_field h_fields1[Nfields1] = {
    {" user",       GTK_JUSTIFY_LEFT,  NULL},
    {" command",    GTK_JUSTIFY_LEFT,  NULL},
    {"pid ",        GTK_JUSTIFY_RIGHT, NULL},
    {"nice ",       GTK_JUSTIFY_RIGHT, NULL},
    {"state",       GTK_JUSTIFY_CENTER,
     "S = sleeping, R = running"},
    {"size ",       GTK_JUSTIFY_RIGHT,
    "total process size in Kbyte"},
    {"rss ",        GTK_JUSTIFY_RIGHT,
    "resident process size in Kbyte"},
    {"time ",       GTK_JUSTIFY_RIGHT,
    "total CPU time of process"},
    {"%CPU",        GTK_JUSTIFY_RIGHT,  NULL},
    {"", GTK_JUSTIFY_CENTER, NULL}
};

static struct header_field h_fields2[Nfields2] = {
    {" hostname",   GTK_JUSTIFY_LEFT,  NULL},
    {" command",    GTK_JUSTIFY_LEFT,  NULL},
    {"pid ",        GTK_JUSTIFY_RIGHT, NULL},
    {"nice ",       GTK_JUSTIFY_RIGHT, NULL},
    {"state",       GTK_JUSTIFY_CENTER,
     "S = sleeping, R = running"},
    {"size ",       GTK_JUSTIFY_RIGHT,
    "total process size in Kbyte"},
    {"rss ",        GTK_JUSTIFY_RIGHT,
    "resident process size in Kbyte"},
    {"time ",       GTK_JUSTIFY_RIGHT,
    "total CPU time of process"},
    {"%CPU",        GTK_JUSTIFY_RIGHT,  NULL},
    {"", GTK_JUSTIFY_CENTER, NULL}
};


/* internet addresses */
static char * addresses[NHOSTS];

/* the table */
static GtkWidget *pages[NHOSTS+2];
static GtkWidget *nb, *label2;

static struct h_tab h[NHOSTS];
static int Index;

/* storage for the row entries */
static gchar * txt[NfieldsMax];
static gchar TxT[NfieldsMax * 24];

static GdkColor bg, bg1;

/*-----------------------------------------------------------------------*/

gint main (int argc, gchar *argv[])
{

    GtkWidget    *window, *window2;
    GtkWidget    *vbox;
    GtkWidget    *hbox;
    GtkWidget    *button;
    GtkWidget    *scrollwin, *label, *list;
    GtkStyle     *style;
    GtkTooltips  *tt;

    char namebuf[20];
    int i, j, k, answer;
    struct hostent * hent;
    pthread_t tid;
    int n0=0, n1=1;

    gtk_init(&argc, &argv);
    g_thread_init(NULL);

    mutex1 = g_mutex_new ();
    mutex2 = g_mutex_new ();

    myuid = getuid();

    for(i=0; i< NfieldsMax; i++)  txt[i] = &TxT[24*i];

    for(i=0; i< NHOSTS;  i++) {
	hent = gethostbyname(hosts[i]);
	if(hent == NULL) {
	    fprintf(stderr, "%s: Unknown host\n", hosts[i]);
	    exit(1);
	}
	addresses[i] = malloc(hent->h_length);
	memcpy(addresses[i], hent->h_addr_list[0], hent->h_length);
	h_length = hent->h_length;

	if(i==0)
	    fprintf(stderr, "Querying host %s...", hosts[i]);
	else
	    fprintf(stderr, "%s...", hosts[i]);

	answer = query(i, addresses, h);

	Hflags[i] = (answer == 0) ? F_ANSWERED : F_NO_RESPONSE;

	if( !( Hflags[i] & F_NO_RESPONSE) ) {
	    for(j = 0; j < h[i].ptab.no; j++) {
		Ptab[i].p[j].pid    = h[i].ptab.p_info[j].pid;
		Ptab[i].p[j].time   = h[i].ptab.p_info[j].time;
		Ptab[i].p[j].cputime= (h[i].ptab.p_info[j].utime +
				       h[i].ptab.p_info[j].stime);
	    }
	    Ptab[i].no = h[i].ptab.no;
	}
    }
    fprintf(stderr, "\n");
    news = 1;

    window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "Network Monitor");
    gtk_signal_connect(GTK_OBJECT(window), "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    /*
    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);
    */
    gtk_rc_parse(NETMONRC);

    window2=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name(GTK_WIDGET(window2), "myrow");
    style = gtk_widget_get_style(window2);
    bg = style->bg[GTK_STATE_NORMAL];
    gtk_widget_destroy(GTK_WIDGET(window2));

    window2=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name(GTK_WIDGET(window2), "myrowspec");
    style = gtk_widget_get_style(window2);
    bg1 = style->bg[GTK_STATE_NORMAL];
    gtk_widget_destroy(GTK_WIDGET(window2));

    /*   gtk_widget_set_usize(window, 80, 270+NHOSTS*18); */
    gtk_window_set_default_size(GTK_WINDOW(window), 465, 30+NHOSTS*18);

    vbox=gtk_vbox_new(FALSE, 0);
    gtk_container_border_width(GTK_CONTAINER(vbox), 0);
    gtk_container_add(GTK_CONTAINER(window), vbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), hbox,FALSE, FALSE, 0);

    button = gtk_button_new_with_label("Quit ");
    gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    nb = gtk_notebook_new();
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(nb), (gboolean)FALSE);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(nb), (gboolean)FALSE);

    button = gtk_button_new_with_label("hostlist ");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
                       GTK_SIGNAL_FUNC(select_page), (gpointer) &n0);

    button = gtk_button_new_with_label("my procs ");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
                       GTK_SIGNAL_FUNC(select_page), (gpointer) &n1);

    button = gtk_button_new_with_label("tabs on/off ");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button),"clicked",
                       GTK_SIGNAL_FUNC(tabs), NULL);

    label2 = gtk_label_new(" click in table to select a host");
    gtk_box_pack_start(GTK_BOX(hbox), label2, FALSE, FALSE, 0);
    gtk_widget_show(label2);

    for(i=0; i< NHOSTS+2;  i++) {
	struct header_field * hh;
	int N;

	if(i == 0 ) {
	    sprintf(namebuf, "hostlist");
	    N = Nfields;
	    hh = h_fields;
	} else if(i == 1) {
	    sprintf(namebuf, "my procs");
	    N = Nfields2;
	    hh = h_fields2;
	} else {
	    sprintf(namebuf, "%s", hosts[i-2]);
	    N = Nfields1;
	    hh = h_fields1;
	}

	label = gtk_label_new(namebuf);
	gtk_widget_show(label);

	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);

	list = gtk_clist_new(N);
	pages[i] = list;

	for(k = 0; k < N; k++) {
	    gtk_clist_set_column_title(GTK_CLIST (list), k, hh[k].name);
	    gtk_clist_set_column_justification(GTK_CLIST(list), k,
					       hh[k].Justification);
	    gtk_clist_set_column_auto_resize (GTK_CLIST (list), k, TRUE);

	    if(h_fields[k].tooltip != NULL) {
		tt = gtk_tooltips_new();
		gtk_tooltips_set_tip(tt,
			    GTK_WIDGET(GTK_CLIST(list)->column[k].button),
				     hh[k].tooltip,NULL);
		gtk_tooltips_set_delay(tt, 50);
	    }
	}

	gtk_clist_column_titles_active (GTK_CLIST (list));
	gtk_clist_column_titles_show(GTK_CLIST (list));

	if( i> 1)
	    gtk_clist_set_selection_mode(GTK_CLIST(list),
					 GTK_SELECTION_EXTENDED);
	else {
	    gtk_clist_set_selection_mode(GTK_CLIST(list),GTK_SELECTION_SINGLE);
	    gtk_signal_connect(GTK_OBJECT(list),"select_row",
			       GTK_SIGNAL_FUNC(row_cb),NULL);
	}

	gtk_signal_connect (GTK_OBJECT (list), "click_column",
			    (GtkSignalFunc) clist_click_column, NULL);

	gtk_clist_set_shadow_type(GTK_CLIST (list), GTK_SHADOW_OUT);
	gtk_widget_show (list);

	gtk_container_add (GTK_CONTAINER (scrollwin), list);
	gtk_widget_show(scrollwin);

	gtk_notebook_append_page(GTK_NOTEBOOK(nb), scrollwin, label);
    }

    gtk_notebook_set_scrollable(GTK_NOTEBOOK(nb), TRUE);

    gtk_box_pack_start(GTK_BOX(vbox), nb, TRUE, TRUE, 0);
    gtk_widget_show(hbox);
    gtk_widget_show (list);
    gtk_widget_show(scrollwin);
    gtk_widget_show(nb);
    gtk_widget_show(vbox);
    gtk_widget_show(window);

    gtk_timeout_add(1000, mon_update, NULL);

    pthread_create(&tid, NULL, &ask_clients, NULL);

    gtk_main ();

    return 0;
}

/*-----------------------------------------------------------------------*/

gint mon_update(gpointer dummy) {

    int flag;
    int i, j, k, n, cpup, m, lno;
    double load;
    long ptime;
    char data[16];
    struct passwd * pw;
    gint old_hoffset, old_voffset;
    float old_adj_hval, old_adj_vval;
    GtkWidget    *list;
    static GtkStyle *StL[NHOSTS]= {NULL};
    GdkColor col1;
    double gr, rd;


    if (! StL[0])    {
	col1.red   = 0;
	col1.green = 0;
	col1.blue  = 0;

	for(i=0; i<NHOSTS;i++) {
	    StL[i] = gtk_style_copy (GTK_WIDGET (pages[0])->style);
	    StL[i]->base[GTK_STATE_NORMAL] = col1;
	}
    }

    g_mutex_lock(mutex2);
    flag = (news == 1 ? 1 : 0);
    news = 0;
    g_mutex_unlock(mutex2);

    if(flag == 1) {
	for(m=0; m<NHOSTS+2; m++) {
	    list = pages[m];
	    old_adj_hval = GTK_CLIST (list)->hadjustment->value;
	    old_adj_vval = GTK_CLIST (list)->vadjustment->value;
	    old_hoffset = GTK_CLIST (list)->hoffset;
	    old_voffset = GTK_CLIST (list)->voffset;

	    gtk_clist_freeze(GTK_CLIST (list));
	    gtk_clist_clear(GTK_CLIST (list));

	    if(m==0) {
		for(i = 0; i < NHOSTS; i++) {
		    sprintf(txt[0]," %s", hosts[i]);
		    if(Hflags[i] & F_NO_RESPONSE) {
			sprintf(txt[1],"dead");
			for(n=2; n< Nfields; n++) *txt[n]=0;
			h[i].ptab.no = 0;
			Ptab[i].no=0;
		    } else {
			sprintf(txt[1],"%i %s %i", h[i].hw.proc_number,
				h[i].hw.processor_name,
				h[i].hw.bogo);
			sprintf(txt[2],"%li", h[i].hw.ram_size);

			sprint_time_val(data, h[i].ls.uptime);
			sprintf(txt[3], "%s", data);
			sprintf(txt[4],"%2.2f", h[i].ls.av1);
			*txt[5] = 0;
		    }
		    lno = gtk_clist_append(GTK_CLIST (list), txt);
		    gtk_clist_set_background(GTK_CLIST(list),lno,&bg);
		    load = h[i].ls.av1 /  h[i].hw.proc_number;


		    load = load < 2. ? load : 2.;

		    if(load<1.) {
			rd = sqrt(load);
			gr = 1;
		    } else {
			rd = 1;
			gr = (2-load)*(2-load);
		    }
		    
		    col1.red   = 65000* rd;
		    col1.green = 65000* gr;
		    col1.blue = 0;
		    if(Hflags[i] & F_NO_RESPONSE) {
			col1.red   = 0;
			col1.green = 0;
			col1.blue =  0;
		    }

		    StL[i]->base[GTK_STATE_NORMAL] = col1;
		    gtk_clist_set_cell_style(GTK_CLIST(list),lno,4,StL[i]);
		}

		sort_procs();

	    } else if (m == 1) {
		int ii;
		for(ii=0; ii<NHOSTS; ii++)
		    for(k = 0; k < h[ii].ptab.no; k++) {
			j = procno[ii][k];
			if( h[ii].ptab.p_info[j].uid == myuid) {
			    sprintf(txt[0]," %s", hosts[ii]);
			    sprintf(txt[1], "%s",  h[ii].ptab.p_info[j].cmd);
			    sprintf(txt[2], "%5d", h[ii].ptab.p_info[j].pid);
			    sprintf(txt[3], "%2ld", h[ii].ptab.p_info[j].nice);
			    sprintf(txt[4], "%c" , h[ii].ptab.p_info[j].state);
			    sprintf(txt[5],"%ld",
				    h[ii].ptab.p_info[j].vsize/1024);
			    sprintf(txt[6], "%ld",
				    h[ii].ptab.p_info[j].rss*PAGE_SIZE/1024);
			    ptime = (h[ii].ptab.p_info[j].utime +
				     h[ii].ptab.p_info[j].stime);
			    sprint_time_val(data, ptime/HZ);
			    sprintf(txt[7], "%s", data);
			    cpup = pcpu[ii][j];
			    if( cpup != -1)
				sprintf(txt[8], "%2d.%d",cpup/10,
					(unsigned) cpup%10 );
			    else
				txt[8][0] = 0;
			    *txt[9] = 0;

			    lno = gtk_clist_append(GTK_CLIST (list), txt);
			    gtk_clist_set_background(GTK_CLIST(list),lno,&bg);
			}
		    }
	    } else if (m > 1) {
		i = m - 2;
		if(Hflags[i] & F_NO_RESPONSE) {
		    sprintf(txt[0],"dead");
		    for(n = 1; n < Nfields1; n++) *txt[n]=0;
		    gtk_clist_append(GTK_CLIST (list), txt);
		} else {
		    for(k = 0; k < h[i].ptab.no; k++) {
			j = procno[i][k];
			pw = getpwuid( h[i].ptab.p_info[j].uid );
			if(pw == NULL)
			    sprintf(txt[0], "%i", h[i].ptab.p_info[j].uid );
			else
			    sprintf(txt[0], "%s",  pw->pw_name);
			sprintf(txt[1], "%s",  h[i].ptab.p_info[j].cmd);
			sprintf(txt[2], "%5d", h[i].ptab.p_info[j].pid);
			sprintf(txt[3], "%2ld", h[i].ptab.p_info[j].nice);
			sprintf(txt[4], "%c" , h[i].ptab.p_info[j].state);
			sprintf(txt[5], "%ld", h[i].ptab.p_info[j].vsize/1024);
			sprintf(txt[6], "%ld",
				h[i].ptab.p_info[j].rss*PAGE_SIZE/1024);
			ptime = (h[i].ptab.p_info[j].utime +
				 h[i].ptab.p_info[j].stime);
			sprint_time_val(data, ptime/HZ);
			sprintf(txt[7], "%s", data);
			cpup = pcpu[i][j];
			if( cpup != -1)
			    sprintf(txt[8], "%2d.%d",cpup/10,
				    (unsigned) cpup%10 );
			else
			    txt[8][0] = 0;
			*txt[9] = 0;

			lno = gtk_clist_append(GTK_CLIST (list), txt);
			if( h[i].ptab.p_info[j].uid != myuid)
			    gtk_clist_set_background(GTK_CLIST(list),lno,&bg);
			else
			    gtk_clist_set_background(GTK_CLIST(list),lno,&bg1);
		    }

		    for(j = 0; j < h[i].ptab.no; j++) {
			Ptab[i].p[j].pid    = h[i].ptab.p_info[j].pid;
			Ptab[i].p[j].time   = h[i].ptab.p_info[j].time;
			Ptab[i].p[j].cputime= (h[i].ptab.p_info[j].utime +
					       h[i].ptab.p_info[j].stime);
		    }
		    Ptab[i].no = h[i].ptab.no;
		}
	    }
	    /* keep old view position - stolen from GNOME's gtop */
	    GTK_CLIST (list)->hadjustment->value = old_adj_hval;
	    GTK_CLIST (list)->vadjustment->value = old_adj_vval;
	    GTK_CLIST (list)->hoffset = old_hoffset;
	    GTK_CLIST (list)->voffset = old_voffset;

	    gtk_clist_thaw(GTK_CLIST (list));
	}
    }
    return TRUE;
}

/*-----------------------------------------------------------------------*/

int cpu_cmp(const int * a, const int * b) {
    return pcpu[Index][*b] - pcpu[Index][*a];
}

void sort_procs() {
    int i, j, k, ctimedelta=0, cpup, ptime;
    double timedelta;

    for(i=0; i<NHOSTS; i++) {
	for(j = 0; j < h[i].ptab.no; j++) {
	    ptime = (h[i].ptab.p_info[j].utime +
		     h[i].ptab.p_info[j].stime);
	    for(k=0,timedelta=0; k<Ptab[i].no; k++) {
		if( Ptab[i].p[k].pid == h[i].ptab.p_info[j].pid) {
		    timedelta = h[i].ptab.p_info[j].time-
			Ptab[i].p[k].time;
		    ctimedelta = ptime - Ptab[i].p[k].cputime;
		    break;
		}
	    }
	    if(timedelta > 0) {
		cpup =  1000*ctimedelta/HZ;
		cpup /= timedelta;
		if(cpup >= 1000) cpup=999;
	    }  else
		cpup = -1;

	    pcpu[i][j] = cpup;
	    procno[i][j] = j;
	}
	Index = i;
	qsort(procno[i], h[i].ptab.no, sizeof(int), cpu_cmp);
    }
}

/*-----------------------------------------------------------------------*/

void * ask_clients(){

    int i,res;

    GMutex *mutex;
    GCond *cond;
    GTimeVal current_time;

    mutex = g_mutex_new ();
    cond = g_cond_new ();

    for(;;) {
	g_get_current_time (&current_time);
	current_time.tv_sec += (guint) UPDTINTERVALL + 1;
	g_mutex_lock (mutex);
	res=g_cond_timed_wait (cond, mutex, &current_time);
	g_mutex_unlock (mutex);
	g_mutex_lock(mutex1);
	for(i = 0; i < NHOSTS; i++) {
	    res=query(i, addresses,  h);
	    Hflags[i] = 0;
	    Hflags[i] = (res == 0) ? F_ANSWERED : F_NO_RESPONSE;
	}
	g_mutex_unlock(mutex1);
	g_mutex_lock(mutex2);
	news = 1;
	g_mutex_unlock(mutex2);
    }
    return NULL;
}

/*-----------------------------------------------------------------------*/

/* callback functions */

void select_page (GtkWidget * w, int* n){
    gtk_notebook_set_page(GTK_NOTEBOOK(nb), *n);
    gtk_label_set_text(GTK_LABEL(label2)," click in table to select a host");
}

/*-----------------------------------------------------------------------*/

void tabs(GtkWidget * w, gpointer data){
    gint tval=FALSE;
    gint bval=FALSE;
    if(GTK_NOTEBOOK(nb)->show_tabs == 0)   tval = TRUE;
    if(GTK_NOTEBOOK(nb)->show_border == 0) bval = TRUE;
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(nb), tval);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(nb), bval);
}

/*-----------------------------------------------------------------------*/

void clist_click_column (GtkCList *clist, gint column, gpointer data)
{

    if (column == clist->sort_column) 	{
	if (clist->sort_type == GTK_SORT_ASCENDING)
	    clist->sort_type = GTK_SORT_DESCENDING;
	else
	    clist->sort_type = GTK_SORT_ASCENDING;
    } else
	gtk_clist_set_sort_column (clist, column);

    gtk_clist_sort (clist);
}

/*-----------------------------------------------------------------------*/

void row_cb(GtkWidget * w, gint row, gint column,
	    GdkEventButton *event, gpointer data){
    gchar *text;
    int i;

    gtk_clist_get_text(GTK_CLIST(w), row, 0, &text);
    gtk_clist_unselect_row(GTK_CLIST(w), row, column);
    gtk_label_set_text(GTK_LABEL(label2),text);
    text++;
    for(i=0; i<NHOSTS; i++)
	if(!strncmp(hosts[i], text, strlen(hosts[i])))
	    gtk_notebook_set_page(GTK_NOTEBOOK(nb),i+2);
}

/*-----------------------------EOF-------------------------------------*/
