#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <asm/param.h>  /* HZ */

#include "config.h"
#include "global.h"

/*-----------------------------------------------------------------------*/


void * net_cat(void * arg);
void * server();
void * write_info();

void get_arch();

void openproc();
void closeproc();
int  get_proc();



/*-----------------------------------------------------------------------*/

static struct proc_info p;
static DIR * procfs;
static struct timeval Time;

static struct h_tab ht;

pthread_mutex_t lock   = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond    = PTHREAD_COND_INITIALIZER;
pthread_mutex_t m_cond = PTHREAD_MUTEX_INITIALIZER;


/*-----------------------------------------------------------------------*/

int main() {
    int i;
    pthread_t tid, tid2;

    /* become a daemon */

    if(  fork()  != 0 ) exit(0);
    setsid();
    signal(SIGHUP, SIG_IGN);
    if(  fork()  != 0 ) exit(0);
    chdir("/");
    for(i=0; i<3; i++) close(i);

    get_arch();

    pthread_create(&tid, NULL, &server, NULL);
    pthread_create(&tid2, NULL, &write_info, NULL);

    pthread_join(tid, NULL);
    pthread_join(tid2, NULL);

    return 0;
}

/*-----------------------------------------------------------------------*/

void * write_info() {
    FILE * P;
    double uptime_sec;
    int n,tim;
    struct timeval tv;
    struct timespec ts;


    for(;;) {
	gettimeofday(&tv, NULL);
	ts.tv_sec = tv.tv_sec + UPDTINTERVALL;
	pthread_mutex_lock(&m_cond);
	pthread_cond_timedwait(&cond, &m_cond, &ts);
	pthread_mutex_unlock(&m_cond);

	pthread_mutex_lock(&lock);

	P = fopen("/proc/uptime","r");
	fscanf(P, "%lf", &uptime_sec);
	ht.ls.uptime = (time_t) uptime_sec;
	fclose(P);
	P = fopen("/proc/loadavg","r");
	fscanf(P, "%lf %lf %lf", &ht.ls.av1, &ht.ls.av5, &ht.ls.av15);
	fclose(P);

	n = 0;
	openproc();
	while( get_proc() == 0 && n < MAX_PROCS_PER_HOST) {
	    tim =  (p.utime + p.stime)/HZ;
	    if ( tim    >= TIME_MIN * 60 ||
		 p.vsize >= SIZE_MIN * MEGA ) {
		memcpy(&ht.ptab.p_info[n], &p,
		       sizeof(struct proc_info));
		n++;
	    }
	}
	closeproc();
	ht.ptab.no = n;

	pthread_mutex_unlock(&lock);
    }
    return NULL;
}

/*-----------------------------------------------------------------------*/

void * server(){

    int listenfd, connfd;
    struct sockaddr_in servaddr, cliaddr;
    pthread_t tid;
    socklen_t addrlen, len;
    const int e = 1;

    listenfd = socket(AF_INET, SOCK_STREAM, 0);
    setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &e, sizeof e);
    memset(&servaddr, 0, sizeof(servaddr));
    addrlen=sizeof(cliaddr);

    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(PORT);


    bind(listenfd, &servaddr, sizeof(servaddr));

    listen(listenfd, 5);

    for(;;) {
	len = addrlen;
	connfd = accept(listenfd, &cliaddr, &len);
	if(connfd < 0) continue;
#ifdef CHECK_NETMASK
	if( memcmp(inet_ntoa(cliaddr.sin_addr), NETMASK,
		   strlen(NETMASK)) != 0) { close(connfd); continue; }
#endif
	pthread_create(&tid, NULL, &net_cat, (void *) connfd);
    }
}


/*-----------------------------------------------------------------------*/

void * net_cat(void * arg) {
    size_t nleft, nwritten;
    char * ptr;
#ifdef CHECK_PAROLE
    unsigned cmd;
#endif
    pthread_detach(pthread_self());

#ifdef CHECK_PAROLE
    if(read((int)arg, &cmd, 4) == 4)
	if(cmd == PAROLE) {
#endif

    pthread_mutex_lock(&lock);
    ptr = (char*)&ht;
    nleft = sizeof(ht);
    while (nleft > 0) {
	if( (nwritten=write( (int) arg, ptr, nleft)) < 0) {
	    if(errno == EINTR) nwritten = 0;
	    else return NULL;
	}
	nleft -= nwritten;
	ptr += nwritten;
    }
    pthread_mutex_unlock(&lock);

#ifdef CHECK_PAROLE
    }
#endif

    close((int) arg);
    return NULL;
}


/*-----------------------------------------------------------------------*/


void get_arch()
{
    FILE *f;
    char buf[256], buf2[70], buf3[70];
    struct stat st;
    unsigned bogo = 0;

    *buf2 = 0;
    *buf3 = 0;

    if((f = fopen("/proc/cpuinfo", "r")) != NULL) {
	while(fgets(buf, 256, f) != NULL) {
	    if(strstr(buf, "processor") == buf){ht.hw.proc_number++; continue;}
	    if(sscanf(buf, "model : %69[a-zA-Z0-9/+ ]", buf2) == 1) continue;
	    if(sscanf(buf, "model name : %69[a-zA-Z0-9/+() ]", buf3) == 1)
		continue;
	    if(sscanf(buf, "bogomips : %u", &bogo) == 1) continue;
	}
	fclose(f);
    }
    buf2[69] = 0;

    if(!stat("/proc/kcore", &st)) ht.hw.ram_size = st.st_size >> 20;

    if(     !strncmp(buf2,"5",           sizeof "5" - 1)) strcpy(buf2, "PII");
    else if(!strncmp(buf2,"3",           sizeof "3" - 1)) strcpy(buf2, "PII");
    else if(!strncmp(buf2, "DX/4",        sizeof "DX/4" - 1))
	strcpy(buf2, "486DX/4"), bogo *= 2;
    else if(!strncmp(buf2, "Pentium 75",  sizeof "Pentium 75" - 1))
	strcpy(buf2, "P"), bogo = 75;
    else if(!strncmp(buf2, "Pentium 75+", sizeof "Pentium 75+" - 1))
	strcpy(buf2, "P"), bogo = 75;
    else if(!strncmp(buf2, "unknown",     sizeof "unknown" - 1))
	strcpy(buf2, "486"), bogo *= 2;
    else if(!strncmp(buf2, "Pentium Pro", sizeof "Pentium Pro" - 1))
	strcpy(buf2, "PPro");
    else if(!strncmp(buf2, "Pentium II",  sizeof "Pentium II" - 1))
	strcpy(buf2, "PII");

    if(!strncmp(buf3,  "Pentium II (Deschutes)",
		sizeof "Pentium II (Deschutes)" - 1))
	strcpy(buf2, "PII (D)");
    else if(!strncmp(buf3,  "Pentium II (Klamath)",
		     sizeof "Pentium II (Klamath)" - 1))
	strcpy(buf2, "PII (K)");
    else if(!strncmp(buf3,  "Pentium Pro",
		     sizeof "Pentium Pro" - 1))
	strcpy(buf2, "PPro");

    switch(bogo) {
    case  62 ...  70 : bogo =  66; break;
    case  71 ...  79 : bogo =  75; break;
    case  96 ... 104 : bogo = 100; break;
    case 196 ... 204 : bogo = 200; break;
    case 262 ... 270 : bogo = 266; break;
    case 296 ... 304 : bogo = 300; break;
    case 346 ... 354 : bogo = 350; break;
    case 396 ... 404 : bogo = 400; break;
    case 446 ... 454 : bogo = 450; break;
    case 496 ... 504 : bogo = 500; break;
    }

    ht.hw.bogo = bogo;
    snprintf(ht.hw.processor_name, 24, "%s", buf2);
}

/*-----------------------------------------------------------------------*/

/* functions to read process table */

void openproc()  { procfs = opendir("/proc"); }
void closeproc() {          closedir(procfs); }

int get_proc()
{
    static struct dirent * ent;
    static char            path[32], sbuf[512];
    static struct stat     sb;
    int fd, num_read;
    char * tmp;

 next_proc:
    /* get next numeric /proc entry */
    while ((ent = readdir(procfs)) &&
	   (*ent->d_name < '0' || *ent->d_name > '9'))
	;
    /* last entry */
    if (!ent || !ent->d_name) return -1;

    sprintf(path, "/proc/%s/stat", ent->d_name);

    if (stat(path, &sb) == -1)		/* no such dirent (anymore) */
	goto next_proc;
    p.uid = sb.st_uid;			/* get uid */

    if ( (fd = open(path, O_RDONLY, 0)) == -1 ) goto next_proc;
    if ( (num_read = read(fd, sbuf , sizeof(sbuf) - 1)) <= 0 ) goto next_proc;
    sbuf[num_read] = 0;
    close(fd);

    tmp = strrchr(sbuf, ')');	/* split into "PID (cmd" and "<rest>" */
    *tmp = '\0';		/* replace trailing ')' with NUL */
    /* parse these two strings separately, skipping the leading "(". */
    memset(p.cmd, 0, sizeof p.cmd);
    sscanf(sbuf, "%d (%39c", &p.pid, p.cmd);
    sscanf(tmp + 2,			/* skip space after ')' too */
	   "%c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u "
	   "%ld %ld "
	   "%*d %*d %*d "
	   "%ld %*u %*u %*d %lu %lu ",
	   &p.state,
	   &p.utime, &p.stime,
	   &p.nice,
	   &p.vsize, &p.rss);

    gettimeofday (&Time, NULL);
    p.time = Time.tv_sec + (double) (Time.tv_usec) / 1000000.0;
    return 0;

}
/*-----------------------------EOF---------------------------------------*/
